#include <stdlib.h>
#include <stdio.h>
#include <inttypes.h>
#include <limits.h>
#include <string.h>

// from thp_play by thakis

inline void toWORD(uint16_t& w)
{
  uint8_t w1 = w & 0xff;
  uint8_t w2 = w >> 8;
  w = (w1 << 8) | w2;
}

inline void toSHORT(int16_t& w)
{
  toWORD(*(uint16_t*)&w);
}

inline void toDWORD(uint32_t& d)
{
  uint8_t w1 = d & 0xff;
  uint8_t w2 = (d >> 8) & 0xff;
  uint8_t w3 = (d >> 16) & 0xff;
  uint8_t w4 = d >> 24;
  d = (w1 << 24) | (w2 << 16) | (w3 << 8) | w4;
}

inline uint32_t aDWORD(uint32_t d)
{
  toDWORD(d); return d;
}

inline void toFLOAT(float& f)
{ toDWORD(*(uint32_t*)&f); }

struct ThpHeader
{
  char tag[4]; //'THP\0'

  //from monk's thp player:
  uint32_t version; //0x00011000 = 1.1, 0x00010000 = 1.0
  uint32_t maxBufferSize;
  uint32_t maxAudioSamples; //!= 0 if sound is stored in file


  float fps; //usually 29.something (=0x41efc28f) for ntsc
  uint32_t numFrames;
  uint32_t firstFrameSize; //size of first frame

  uint32_t dataSize; //size of file - ThpHeader.offset

  //from monk's thp player:
  uint32_t componentDataOffset; //ThpComponents stored here
  uint32_t offsetsDataOffset; //?? if != 0, offset to table with offsets of all frames?

  uint32_t firstFrameOffset;
  uint32_t lastFrameOffset;
};

//monk:

struct ThpComponents
{
  uint32_t numComponents; //usually 1 or 2 (video or video + audio)

  //component type 0 is video, type 1 is audio,
  //type 0xff is "no component" (numComponent many entries
  //are != 0xff)
  uint8_t componentTypes[16];
};

struct ThpVideoInfo
{
  uint32_t width;
  uint32_t height;
  uint32_t unknown; //only for version 1.1 thp files
};

struct ThpAudioInfo
{
  uint32_t numChannels;
  uint32_t frequency;
  uint32_t numSamples;
  uint32_t numData; //only for version 1.1 - that many
               //audio blocks are after each video block
               //(for surround sound?)
};

struct ThpAudioBlockHeader
{
  //size 80 byte
  uint32_t channelSize; //size of one channel in bytes
  uint32_t numSamples; //number of samples/channel
  int16_t table1[16]; //table for first channel
  int16_t table2[16]; //table for second channel
  int16_t channel1Prev1;
  int16_t channel1Prev2;
  int16_t channel2Prev1;
  int16_t channel2Prev2;
};

ThpHeader			_Head;
ThpComponents		_Components;
ThpVideoInfo		_VideoInfo;
ThpAudioInfo		_AudioInfo;
ThpAudioBlockHeader	_AudioBlockHeader;

int					_numInts;

uint32_t			_currFrameNr;


void readThpHeader(FILE *hFile, ThpHeader &h)
{
    size_t readBytes;

    readBytes = fread(&h,1,sizeof(ThpHeader),hFile);

	if(readBytes==sizeof(ThpHeader))
	{
		toDWORD(h.version);  
		toDWORD(h.maxBufferSize);
		toDWORD(h.maxAudioSamples);
		toFLOAT(h.fps);
		toDWORD(h.numFrames);
		toDWORD(h.firstFrameSize);
		toDWORD(h.dataSize);
		toDWORD(h.componentDataOffset);
		toDWORD(h.offsetsDataOffset);
		toDWORD(h.firstFrameOffset);
		toDWORD(h.lastFrameOffset);
	}
}

void readThpComponents(FILE *hFile, ThpComponents &c)
{
    fread(&c,1,sizeof(ThpComponents),hFile);

	toDWORD(c.numComponents);
}

void readThpVideoInfo(FILE *hFile, ThpVideoInfo& i, bool isVersion11)
{
    fread(&i,1,sizeof(ThpVideoInfo),hFile);

	toDWORD(i.width);
	toDWORD(i.height);
	if(isVersion11)
		toDWORD(i.unknown);
	else
	{
		i.unknown = 0;
        fseek(hFile,-4,SEEK_CUR);
	}
}

void readThpAudioInfo(FILE *hFile, ThpAudioInfo& i, bool isVersion11)
{
    fread(&i,1,sizeof(ThpAudioInfo),hFile);

	toDWORD(i.numChannels);
	toDWORD(i.frequency);
	toDWORD(i.numSamples);
	
	if(isVersion11)
		toDWORD(i.numData);
	else
	{
	    i.numData = 1;
        fseek(hFile,-4,SEEK_CUR);
	}	
}

int main(int argc, char* argv[])
{

	FILE    *hFile;
    FILE    *hFileSave;

	uint32_t		currAudioOffset;
	uint32_t		nextAudioOffset;
	uint32_t		offsetValue;

	uint32_t		currAudioLength;
	uint32_t		nextAudioLength;

	uint32_t		currFrameOffset=0;
	uint32_t		currFrameSize=0;
	uint32_t		nextFrameOffset=0;
	uint32_t		nextFrameSize=0;

	static char	filename[PATH_MAX+1];
	static char	fileout[PATH_MAX+1];

	bool	bSaveHeader=0;
	uint8_t	Header[0x60];

	unsigned char *	_currFrameData;
	unsigned char *	_nextFrameData;
	
	if (argc!=2)
	{
		fprintf(stderr,"Usage : %s <filename>\n",argv[0]);
        exit(1);
	}

	strncpy(filename,argv[1],PATH_MAX);
	strncpy(fileout,argv[1],PATH_MAX);

    hFile=fopen(filename,"rb");
	if(hFile==NULL)
	{
		fprintf(stderr,"Unable to open %s !\n",argv[1]);
        exit(1);
	}
	
	// Stock the THP Header
    fread(Header,1,0x60,hFile);
    fseek(hFile,0,SEEK_SET);

	// Check if THP :P
	if(memcmp(Header,"THP\0x00",4))
	{
		fprintf(stdout,"%s is not a THP file ! \n",argv[1]);
        exit(1);
	}

	// Read Stuff ..
	readThpHeader(hFile,_Head);
	readThpComponents(hFile, _Components);

	for(uint32_t i = 0; i < _Components.numComponents; ++i)
	{
		if(_Components.componentTypes[i] == 0) //video
			readThpVideoInfo(hFile, _VideoInfo, _Head.version == 0x00011000);
		else if(_Components.componentTypes[i] == 1) //audio
		{
			readThpAudioInfo(hFile, _AudioInfo, _Head.version == 0x00011000);
		}
	}

	// Check if file has audio ...
	if(_Head.maxAudioSamples==0)
	{
		fprintf(stdout,"%s has no audio ! \n",argv[1]);
        exit(1);
	}

	memcpy(&fileout[strlen(fileout)-3],"DSP",3);

    hFileSave=fopen(fileout,"wb");
	if(hFileSave==NULL)
	{
		fprintf(stderr,"Unable to open save file !\n",argv[1]);
        exit(1);
	}

	_currFrameNr = -1;
	nextFrameOffset = _Head.firstFrameOffset;
	nextFrameSize = _Head.firstFrameSize;

	_currFrameData = (unsigned char*)malloc(_Head.maxBufferSize);
	_nextFrameData = (unsigned char*)malloc(_Head.maxBufferSize);

	if(!(_currFrameData || _nextFrameData))
	{
		fprintf(stderr,"Unable to allocate %d bytes ! \n",_Head.maxBufferSize);
        exit(1);
	}
	
	// Display file info ...
	fprintf(stdout,"File : %s \n",filename);
	fprintf(stdout,"Out  : %s \n",fileout);
	if(_AudioInfo.numChannels==1)
		fprintf(stdout,"%d Hz mono\n\n",_AudioInfo.frequency);
	else
		fprintf(stdout,"%d Hz stereo\n\n",_AudioInfo.frequency);

	currFrameOffset=nextFrameOffset;
	
	do
	{
		++_currFrameNr;

		// Get the size of the first chunk ...
        fseek(hFile,nextFrameOffset,SEEK_SET);
        fread(_currFrameData,1,nextFrameSize,hFile);
		
		currAudioOffset = aDWORD(*(uint32_t*)(_currFrameData+0x08))+0x10;
		currAudioLength=aDWORD(*(uint32_t*)(_currFrameData+currAudioOffset));
		currAudioLength*=_AudioInfo.numChannels;
		currAudioLength+=sizeof(ThpAudioBlockHeader)+0x010;

		currFrameOffset += nextFrameSize;
		currFrameSize = aDWORD(*(uint32_t*)_currFrameData);
	
		// Get the size of the second chunk ...
        fseek(hFile,currFrameOffset,SEEK_SET);
        fread(_nextFrameData,1,currFrameSize,hFile);

		nextAudioOffset = aDWORD(*(uint32_t*)(_nextFrameData+0x08))+0x10;
		nextAudioLength = aDWORD(*(uint32_t*)(_nextFrameData+nextAudioOffset));
		nextAudioLength *= _AudioInfo.numChannels;
		nextAudioLength+=sizeof(ThpAudioBlockHeader)+0x10;

		if(!bSaveHeader)
		{
			// Save the THPHeader
			bSaveHeader=1;
			offsetValue=aDWORD(currAudioLength);
		    memcpy(Header+0x18,&offsetValue,4);

            fwrite(Header,1,0x60,hFileSave);
		}

		// Write Offsets ...
		offsetValue=aDWORD(nextAudioLength);
        fwrite(&offsetValue,1,4,hFileSave);
        fwrite(&offsetValue,1,4,hFileSave);
		offsetValue=0;
        fwrite(&offsetValue,1,4,hFileSave);
        fwrite(&offsetValue,1,4,hFileSave);

		// Write Data
        fwrite(_currFrameData+currAudioOffset,1,currAudioLength-0x10,hFileSave);
		
		nextFrameOffset=currFrameOffset;
		nextFrameSize=currFrameSize;

		printf("Processing Frame %d on %d ...\r",_currFrameNr, _Head.numFrames+1);
		
	} while(_currFrameNr <= _Head.numFrames);
	
    fclose(hFileSave);
    fclose(hFile);
	free(_currFrameData);
	free(_nextFrameData);

	return 0;
}
